PlaceableHusbandryAnimalsExtension = {}

PlaceableHusbandryAnimalsExtension.HEALTH_REDUCING_WITH_OVERCROWDING = 25
PlaceableHusbandryAnimalsExtension.DEATH_RATE_WITHOUT_HEALTH = 0.1

function PlaceableHusbandryAnimalsExtension.registerFunctions(placeableType, superFunc)
	superFunc(placeableType)
	SpecializationUtil.registerFunction(placeableType, "calculateMaleFemaleAnimals", PlaceableHusbandryAnimals.calculateMaleFemaleAnimals)
	SpecializationUtil.registerFunction(placeableType, "setMonthSinceBirthIfNeeded", PlaceableHusbandryAnimals.setMonthSinceBirthIfNeeded)
end

PlaceableHusbandryAnimals.registerFunctions = Utils.overwrittenFunction(PlaceableHusbandryAnimals.registerFunctions, PlaceableHusbandryAnimalsExtension.registerFunctions)

function PlaceableHusbandryAnimalsExtension:addAnimals(superFunc, subTypeIndex, numAnimals, age)
    local cluster = g_currentMission.animalSystem:createClusterFromSubTypeIndex(subTypeIndex)

	if cluster:getSupportsMerging() then
		cluster.numAnimals = numAnimals
		cluster.age = age
		cluster.subTypeIndex = subTypeIndex
        self:setMonthSinceBirthIfNeeded(subTypeIndex, cluster)

		self:addCluster(cluster)
	else
		for i = 1, numAnimals do
			cluster = g_currentMission.animalSystem:createClusterFromSubTypeIndex(subTypeIndex)
			cluster.numAnimals = 1
			cluster.age = age
            self:setMonthSinceBirthIfNeeded(subTypeIndex, cluster)

			self:addCluster(cluster)
		end
	end
end

PlaceableHusbandryAnimals.addAnimals = Utils.overwrittenFunction(PlaceableHusbandryAnimals.addAnimals, PlaceableHusbandryAnimalsExtension.addAnimals)

function PlaceableHusbandryAnimalsExtension:updateOutput(superFunc2, superFunc, cluster, foodFactor, productionFactor, clusterProductionFactor)
	-- TerraLiveExtension.DebugText("PlaceableHusbandryAnimalsExtension:updateOutput(%s, %s, %s, %s, %s, %s)", superFunc2, superFunc, cluster, foodFactor, productionFactor, clusterProductionFactor);
	if self.isServer then
		local spec = self.spec_husbandryAnimals

        if spec:getNumOfAnimals() > spec:getMaxNumOfAnimals() then
            local subType = g_currentMission.animalSystem:getSubTypeByIndex(cluster:getSubTypeIndex())
            if foodFactor > subType.healthThresholdFactor then
                foodFactor = subType.healthThresholdFactor
            end
        end

		cluster:updateHealth(foodFactor)

		self:raiseActive()
	end

	superFunc(self, cluster, foodFactor, productionFactor, clusterProductionFactor)
end

PlaceableHusbandryAnimals.updateOutput = Utils.overwrittenFunction(PlaceableHusbandryAnimals.updateOutput, PlaceableHusbandryAnimalsExtension.updateOutput)

function PlaceableHusbandryAnimalsExtension:getConditionInfos(superFunc2, superFunc)
	local infos = superFunc(self)
	-- local spec = self.spec_husbandryAnimals
	-- local animalTypeIndex = self:getAnimalTypeIndex()

	-- if animalTypeIndex ~= AnimalType.HORSE and animalTypeIndex ~= AnimalType.PIG then
		-- local productivity = self:getGlobalProductionFactor()
		-- spec.info.value = productivity
		-- spec.info.ratio = productivity
		-- spec.info.valueText = string.format("%d %%", g_i18n:formatNumber(productivity * 100, 0))

		-- table.insert(infos, spec.info)
	-- end

	return infos
end
PlaceableHusbandryAnimals.getConditionInfos = Utils.overwrittenFunction(PlaceableHusbandryAnimals.getConditionInfos, PlaceableHusbandryAnimalsExtension.getConditionInfos)

function PlaceableHusbandryAnimalsExtension:onPeriodChanged(superFunc2, superFunc)
	if self.isServer then
		local spec = self.spec_husbandryAnimals
		local clusters = spec.clusterSystem:getClusters()
        local totalNewAnimals = 0

		for _, cluster in ipairs(clusters) do
			cluster:onPeriodChanged()

			local numNewAnimals = cluster:updateReproduction()
            totalNewAnimals = totalNewAnimals + numNewAnimals

			if numNewAnimals > 0 then
                local numOfFemaleAnimals, numOfMaleAnimals = self:calculateMaleFemaleAnimals(numNewAnimals)

                local femaleSubType = g_currentMission.animalSystem:getSubTypeByIndex(cluster:getSubTypeIndex())

                if femaleSubType.malePart == nil then
                    numOfFemaleAnimals = numOfFemaleAnimals + numOfMaleAnimals
                    numOfMaleAnimals = 0
                end

                if numOfFemaleAnimals > 0 then
                    local newFemaleCluster = g_currentMission.animalSystem:createClusterFromSubTypeIndex(cluster:getSubTypeIndex())
                    newFemaleCluster.numAnimals = numOfFemaleAnimals
                    newFemaleCluster.health = 100
                    spec.clusterSystem:addPendingAddCluster(newFemaleCluster)
                end

                if numOfMaleAnimals > 0 then
                    local newMaleCluster = g_currentMission.animalSystem:createClusterFromSubTypeIndex(cluster:getMaleSubTypeIndex())
                    newMaleCluster.numAnimals = numOfMaleAnimals
                    newMaleCluster.health = 100
                    spec.clusterSystem:addPendingAddCluster(newMaleCluster)
                end

                local subType = g_currentMission.animalSystem:getSubTypeByIndex(cluster:getSubTypeIndex())
                local animalType = g_currentMission.animalSystem:getTypeByIndex(subType.typeIndex)

                if animalType.statsBreedingName ~= nil then
                    local stats = g_currentMission:farmStats(self:getOwnerFarmId())
                    stats:updateStats(animalType.statsBreedingName, numNewAnimals)
                end
			end
		end

        local totalAnimals = self:getNumOfAnimals() + totalNewAnimals

        if totalAnimals > spec.maxNumAnimals then
            local msg = string.format(g_i18n:getText("TL_husbandry_overcrowded"), self:getName())
            g_client:getServerConnection():sendEvent(InGameNotification.new(InGameNotification.TYPE.CRITICAL, msg, self:getOwnerFarmId()))
        end

		self:raiseActive()
	end
end

PlaceableHusbandryAnimals.onPeriodChanged = Utils.overwrittenFunction(PlaceableHusbandryAnimals.onPeriodChanged, PlaceableHusbandryAnimalsExtension.onPeriodChanged)

function PlaceableHusbandryAnimalsExtension:onDayChanged(superFunc)
	superFunc(self)

    if not self.isServer then
        return
    end
		
    local spec = self.spec_husbandryAnimals
	local clusters = spec.clusterSystem:getClusters()

    for _, cluster in ipairs(clusters) do
        if cluster:getHealthFactor() == 0 and self:getOwnerFarmId() ~= 0 then
            local maxDeadAnimals = math.floor(cluster:getNumAnimals() * PlaceableHusbandryAnimalsExtension.DEATH_RATE_WITHOUT_HEALTH)
            local numOfDeadAnimals = math.random(0, maxDeadAnimals)
            cluster:changeNumAnimals(-numOfDeadAnimals)

            if numOfDeadAnimals > 0 then
                local msg = string.format(g_i18n:getText("TL_husbandry_dec_by_health_single"), self:getName(), numOfDeadAnimals)
                if numOfDeadAnimals > 1 then
                    msg = string.format(g_i18n:getText("TL_husbandry_dec_by_health_multiple"), self:getName(), numOfDeadAnimals)
                end
                g_client:getServerConnection():sendEvent(InGameNotification.new(InGameNotification.TYPE.CRITICAL, msg, self:getOwnerFarmId()))
            end
        end

        if spec:getNumOfAnimals() > spec:getMaxNumOfAnimals() then
            cluster:changeHealth(-PlaceableHusbandryAnimalsExtension.HEALTH_REDUCING_WITH_OVERCROWDING)
        end
    end

	self:raiseActive()
end

PlaceableHusbandryAnimals.onDayChanged = Utils.overwrittenFunction(PlaceableHusbandryAnimals.onDayChanged, PlaceableHusbandryAnimalsExtension.onDayChanged)

--------------------------------------------------- New PlaceableHusbandryAnimals functions ---------------------------------------------------

function PlaceableHusbandryAnimals:calculateMaleFemaleAnimals(numNewAnimals)
    local numOfMaleAnimals = 0
    local numOfFemaleAnimals = 0

    for i=1, numNewAnimals, 1 do
        local randomValue = math.random(2)

        if randomValue == 1 then
            numOfFemaleAnimals = numOfFemaleAnimals + 1
        else
            numOfMaleAnimals = numOfMaleAnimals + 1
        end
    end

    return numOfFemaleAnimals, numOfMaleAnimals
end

function PlaceableHusbandryAnimals:setMonthSinceBirthIfNeeded(subTypeIndex, cluster)
    local subType = g_currentMission.animalSystem:getSubTypeByIndex(subTypeIndex)
    local neededAge = subType.reproductionMinAgeMonth + subType.reproductionDurationMonth
    if subType.supportsReproduction and neededAge <= cluster:getAge() and cluster.monthSinceLastBirth == nil then
        cluster.monthSinceLastBirth = cluster:getAge() - neededAge
    end
end